<?php

namespace Drupal\content_first\Controller;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Component\Utility\DeprecationHelper;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use League\HTMLToMarkdown\HtmlConverter;
use League\HTMLToMarkdown\Converter\TableConverter;

/**
 * Returns Node routes.
 */
class NodeController extends EntityController {

  /**
   * Render Output of a node.
   *
   * @param \Drupal\Core\Entity\EntityInterface $node
   *
   * @return mixed|\Drupal\Component\Render\MarkupInterface
   */
  public function nodeRender($node) {
    if (!$node instanceof EntityInterface) {
      throw new NotFoundHttpException();
    }
    $view_mode = 'full';
    $view_builder = $this->entityTypeManager->getViewBuilder($node->getEntityTypeId());
    $pre_render = $view_builder->view($node, $view_mode);
    $render_output = DeprecationHelper::backwardsCompatibleCall(
     currentVersion: \Drupal::VERSION,
     deprecatedVersion: '10.3',
     currentCallable: fn() => $this->renderer->renderInIsolation($pre_render),
     deprecatedCallable: fn() => $this->renderer->renderPlain($pre_render),
    );
    return $render_output;
  }

  /**
   * Builds the response.
   */
  public function build($node) {

    $build = [];
    if (!$node instanceof EntityInterface) {
      throw new NotFoundHttpException();
    }

    // Render entity HTML output:
    $render_output = $this->nodeRender($node);
    if (empty($render_output)) {
      return NULL;
    }

    $converter = new HtmlConverter(
      [
        'strip_tags' => TRUE,
        // Needed h1:# and H2:# style headers for JS marked plugin.
        'header_style' => 'atx',
      ]
    );

    // Generate the content:
    // Limit output to specific tags and attributes:
    $allowed_tags = '<h1><h2><h3><h4><h5><h6>';
    $allowed_tags .= '<a><p><br>';
    $allowed_tags .= '<strong><em><b><i><u>';
    $allowed_tags .= '<ul><li><ol>';
    $allowed_tags .= '<table><tr><td><th><tbody><thead><tfoot>';
    // Media & embedding tags:
    $allowed_tags .= '<img><embed>';
    $allowed_tags .= '<video><audio><iframe><object><canvas><svg>';
    $render_output = strip_tags($render_output, $allowed_tags);

    // Transform not supported Markdown tags so they are still visible:
    $render_output = $this->transformSelfClosingTags($render_output, ['embed'], ['alt', 'title']);
    $render_output = $this->transformPairedTags($render_output, ['video', 'audio', 'iframe', 'object', 'canvas', 'svg'], ['alt', 'title']);

    // Support for tables:
    $converter->getEnvironment()->addConverter(new TableConverter());
    $markdown = $converter->convert((string) $render_output);
    // Remove extra line breaks:
    $markdown = preg_replace('/(\n\s*){2,}/', "\n\n", $markdown);

    $dom = new \DOMDocument();
    // Suppress warnings due to malformed HTML.
    $html = mb_encode_numericentity($render_output, [0x80, 0x10FFFF, 0, 0x1FFFFF], 'UTF-8');
    @$dom->loadHTML($html);
    $xpath = new \DOMXPath($dom);

    if (!$xpath instanceof \DOMXPath) {
      return NULL;
    }

    // Generate headers menu items:
    $headers_items = [];
    // Query all h1 to h6 tags.
    $headers = $xpath->query('//h1 | //h2 | //h3 | //h4 | //h5 | //h6');
    if ($headers) {
      $headers_items = $this->buildHeadersMenuItemsFromDom($headers);
    }

    if (!empty($html)) {
      $html = (string) $this->addMarkdownToHeadings($html);
    }

    // Main container:
    $build['container'] = [
      '#type' => 'container',
      '#attributes' => [
        'id' => ['content-first'],
      ],
    ];

    // Content container:
    $build['container']['content'] = [
      '#type' => 'container',
      '#attributes' => [
        'class' => ['content'],
      ],
    ];

    $build['container']['content']['actions'] = [
      '#type' => 'container',
      '#attributes' => [
        'class' => ['actions'],
      ],
    ];

    $build['container']['content']['actions']['view'] = [
      '#type' => 'button',
      '#value' => '🗐 ' . $this->t('HTML'),
      '#attributes' => [
        'id' => ['view-button'],
        'class' => ['action-button'],
      ],
    ];

    $build['container']['content']['actions']['copy'] = [
      '#type' => 'button',
      '#value' => '🗐 ' . $this->t('Copy'),
      '#attributes' => [
        'id' => ['copy-button'],
        'class' => ['action-button'],
      ],
    ];

    $build['container']['content']['actions']['txt'] = [
      '#type' => 'button',
      '#value' => '⤓ ' . $this->t('Download'),
      '#attributes' => [
        'id' => ['txt-button'],
        'class' => ['action-button'],
      ],
    ];

    $build['container']['content']['html'] = [
      '#type' => 'container',
      '#markup' => '<pre>' . $markdown . '</pre>',
      '#attributes' => [
        'id' => ['content-first-content'],
      ],
    ];

    // Headers container:
    $build['container']['headers'] = [
      '#theme' => 'item_list',
      '#items' => $headers_items,
      '#list_type' => 'ul',
      '#attributes' => [
        'class' => ['headers-menu'],
      ],
    ];

    $filename = preg_replace('/[^A-Za-z0-9_\-]/', '_', $node->label() . '--' . $node->toUrl()->toString()) . '.txt';
    $build['#attached'] = [
      'library' => [
        'content_first/content_first',
        'content_first/marked',
      ],
      'drupalSettings' => [
        'content_first' => [
          'label' => $node->label(),
          'url' => $node->toUrl()->toString(),
          'filename' => $filename . '.txt',
        ],
      ],
    ];

    return $build;
  }

  /**
   * Build a multi-level menu items from a DOMNodeList of heading elements.
   *
   * Each DOMNode is expected to be an element like <h2> or <h3>.
   * We interpret the number in the tagName ("H2" => 2, "H3" => 3, etc.)
   * as the nesting level.
   *
   * The returned structure is an array of items, where each item has:
   *   [
   *     '#markup'  => (string) The text inside the heading,
   *     '#level'   => (int)    The heading level,
   *     'children' => array    Nested sub-items,
   *   ]
   *
   * @param \DOMNodeList $nodeList
   *   A DOMNodeList of heading elements (e.g. <h1>, <h2>, <h3>, etc.).
   *
   * @return array
   *   Nested array of items with '#markup', '#level', 'children'.
   */
  public function buildHeadersMenuItemsFromDom(\DOMNodeList $nodeList) {
    // Top-level items in its 'children'.
    $virtualRoot = [
      '#markup'  => '',
      '#level'   => 0,
      'children' => [],
    ];

    // Stack to track current nesting context.
    // Each stack element: [ 'level' => int, 'item' => &theItem ]
    // Start with the virtual root.
    $stack = [
      [
        'level' => 0,
        'item'  => &$virtualRoot,
      ],
    ];

    // Iterate each heading node in the DOMNodeList.
    foreach ($nodeList as $node) {
      if (!$node instanceof \DOMElement) {
        continue;
      }

      // Skip if it doesn't start with 'H'.
      $tagName = strtoupper($node->tagName);
      if (strpos($tagName, 'H') !== 0) {
        continue;
      }

      // Skip or handle non-numeric or invalid headings.
      $lvl = (int) substr($tagName, 1);
      if ($lvl <= 0) {
        continue;
      }

      // The heading text is the node's textContent:
      $text = "<span class=\"html-tag\">$node->tagName</span> $node->textContent";

      // Climb up the stack while the top of stack has a level >= $lvl.
      while (count($stack) > 0 && $stack[count($stack) - 1]['level'] >= $lvl) {
        array_pop($stack);
      }

      // Now the top of stack is a heading whose level < $lvl (or virtual root).
      if (!empty($stack)) {
        $parentItem = &$stack[count($stack) - 1]['item'];

        // Create a new item array.
        $newItem = [
          '#markup'  => $text,
          '#level'   => $lvl,
          'children' => [],
        ];

        // Append it to the parent's 'children'.
        $parentItem['children'][] = $newItem;

        // Push this new item onto the stack so deeper headings become its
        // children.
        $newIndex = array_key_last($parentItem['children']);
        $stack[] = [
          'level' => $lvl,
          'item'  => &$parentItem['children'][$newIndex],
        ];
      }
    }

    // Return everything under the virtual root (the real items).
    return $virtualRoot['children'];
  }

  /**
   * Transform self-closing tags into bracketed tokens.
   *
   * E.g. <img src="foo" alt="bar" /> -> [IMG src="foo" alt="bar"].
   *
   * @param string $content
   *   The HTML to process.
   * @param string[] $tags
   *   Array of tag names, e.g. ['img','br','hr'].
   * @param string[] $attributes
   *   Array of attributes to capture, e.g. ['src','alt','title'].
   *
   * @return string
   *   The processed HTML content.
   */
  public function transformSelfClosingTags(string $content, array $tags, array $attributes): string {
    if (empty($content)) {
      return '';
    }

    // Match <tag ...> or <tag .../> ignoring content.
    $tagPattern = implode('|', array_map('preg_quote', $tags));
    $pattern = '/<(' . $tagPattern . ')\b([^>]*)\/?>/i';

    $content = preg_replace_callback($pattern, function ($matches) use ($attributes) {
      $tagName    = strtolower($matches[1]);
      $attrString = $matches[2] ?? '';

      $collected = [];
      foreach ($attributes as $attr) {
        // Only matching double quotes attributes:
        $attrPattern = '/\b' . preg_quote($attr, '/') . '="([^"]*)"/i';
        if (preg_match($attrPattern, $attrString, $attrMatch)) {
          $collected[] = $attr . '="' . $attrMatch[1] . '"';
        }
      }

      if ($collected) {
        return '[' . strtoupper($tagName) . ' ' . implode(' ', $collected) . ']';
      }
      else {
        return '[' . strtoupper($tagName) . ']';
      }
    }, $content);

    return (empty($content)) ? '' : $content;
  }

  /**
   * Transform paired tags like <iframe ...>...</iframe> into bracketed tokens.
   *
   * E.g.  <iframe title="Title"></iframe> -> [IFRAME title="Title"]
   *
   * @param string $content
   *   The HTML to process.
   * @param string[] $tags
   *   Array of tag names, e.g. ['iframe','video'].
   * @param string[] $attributes
   *   Array of attributes to capture, e.g. ['src','title'].
   *
   * @return string
   *   The processed HTML content.
   */
  public function transformPairedTags(string $content, array $tags, array $attributes): string {
    if (empty($content)) {
      return '';
    }

    $tagPattern = implode('|', array_map('preg_quote', $tags));
    $pattern = '/<(' . $tagPattern . ')\b([^>]*)>([^>]*)<\/\1>/i';

    $content = preg_replace_callback($pattern, function ($matches) use ($attributes) {
      // $matches[1] = the matched tag name (e.g. "iframe")
      // $matches[2] = the attribute string
      // $matches[3] = inner content between <...> and </...> (ignored)
      // Normalize the tag name in uppercase in the final bracket:
      $tagName = strtoupper($matches[1]);

      $attrString = $matches[2] ?? '';
      $collected = [];
      foreach ($attributes as $attr) {
        // Only matching double‐quoted attributes:
        $attrPattern = '/\b' . preg_quote($attr, '/') . '="([^"]*)"/i';
        if (preg_match($attrPattern, $attrString, $m)) {
          $collected[] = $attr . '="' . $m[1] . '"';
        }
      }

      $result = '[' . $tagName;
      if (!empty($collected)) {
        $result .= ' ' . implode(' ', $collected);
      }
      $result .= ']<br />';

      return trim($result);
    }, $content);

    return (empty($content)) ? '' : $content;
  }

  /**
   * Add Markdown-style anchors for.
   *
   * @param string $html
   *   The HTML content to process.
   *
   * @return string|null
   *   The processed HTML content.
   */
  public function addMarkdownToHeadings($html) {
    if (empty($html)) {
      return NULL;
    }
    // Process from h1 to h6.
    for ($i = 1; $i <= 6; $i++) {
      if (empty($html)) {
        return NULL;
      }
      $html = preg_replace_callback(
        '/<h' . $i . '(.*?)>(.*?)<\/h' . $i . '>/is',
        function ($matches) use ($i) {
          $attrs = $matches[1];
          $content = trim($matches[2]);
          $hashes = str_repeat('#', $i);
          return "<h$i$attrs>$hashes $content</h$i>";
        },
        $html
      );
    }
    return $html;
  }

}
